create or replace type tp_pessoa as object(
Nome varchar2(60),
Sexo char,
Cadastro number(10),
Foto BLOB,

member function getNome RETURN varchar2,
member function getSexo RETURN char,
member function getCadastro RETURN number,
member function getFoto RETURN BLOB,
member procedure setCadastro(novoCadastro number),
map member function ordenar RETURN number
)not final not instantiable; 
/

create or replace type body tp_pessoa as
member function getNome return varchar2 is
begin
	return self.Nome;
end;
member function getSexo return char is
begin
	return self.Sexo;
end;
member function getCadastro return number is
begin
	return self.Cadastro;
end;
member function getFoto return blob is
begin
	return self.Foto;
end;
member procedure setCadastro(novoCadastro number) is
begin
	self.cadastro := novoCadastro;
end;
map member function ordenar return number is
c number := cadastro;
begin
	return c;
end;
end;
/

create or replace type tp_telefone as object(
Codigo_Pais Number,
Codigo_Cidade Number,
Numero Number,

member function getCodigoPais return number,
member function getCodigoCidade return number,
member function getNumero return number,
member function toString return varchar2
);
/

create or replace type body tp_telefone as
member function getCodigoPais return number is
begin
	return self.Codigo_Pais;
end;
member function getCodigoCidade return number is
begin
	return self.Codigo_Cidade;
end;
member function getNumero return number is
begin
	return self.Numero;
end;
member function toString return varchar2 is
retorno varchar2(50) := self.Codigo_Pais || ' ' || self.Codigo_Cidade || ' ' || Substr(self.Numero || '', 1, 4) || '-' || Substr(self.Numero || '', 5, 4);
begin
	return retorno;
end;
end;
/

create or replace type tp_va_telefone as VARRAY(5) of tp_telefone;
/

create or replace type tp_endereco as object(
Logradouro Varchar2(50),
Numero Number,
Bairro Varchar2(50),
CEP Varchar2(10),
Complemento Varchar2(50),
Cidade Varchar2(50),
UF Varchar2(50),
Pais Varchar2(50),
member function toString return varchar2 
);
/

create or replace type body tp_endereco as
member function toString return varchar2 is
begin
	return 'Logradouro: ' || self.Logradouro || '\nn ' || self.Numero || '\nBairro: ' ||
	self.Bairro || '\nComplemento: ' || self.Complemento || '\nCidade: ' || self.Cidade ||
	'\nUF: ' || self.UF || '\nPais: ' || self.Pais || '\nCEP: ' || self.CEP;
end;
end;
/

create or replace type tp_quarto;
/

create or replace type tp_nt_ref_quarto as table of ref tp_quarto;
/

create or replace type tp_hospede under tp_pessoa(
Telefones tp_va_telefone,
Endereco tp_endereco,
Nascimento date,

member function getTelefones return tp_va_telefone,
member function getEndereco return tp_endereco,
member function getNascimento return date
); 
/

create or replace type body tp_hospede as
member function getTelefones return tp_va_telefone is
begin
	return self.Telefones;
end;
member function getEndereco return tp_endereco is
begin
	return self.Endereco;
end; 
member function getNascimento return date is
begin
	return self.Nascimento;
end;
end;
/

create table tb_hospede of tp_hospede(
Cadastro primary key
);

create or replace type tp_nt_ref_hospede as table of ref tp_hospede;
/

create or replace type tp_quarto as object(
Categoria varchar2(50),
Numero number,
Capacidade number,

--Cria um quarto no andar especificado
constructor function tp_quarto (numero number, categoria varchar2, capacidade number) return self as result,
order member function ordem(x in tp_quarto) return number 
);
/

create or replace type body tp_quarto as
constructor function tp_quarto (numero number, categoria varchar2, capacidade number) return self as result is
begin
self.Numero := numero;
self.Categoria := categoria;
self.Capacidade := capacidade;
end;
order member function ordem(x in tp_quarto) return number is
begin
	return self.Numero - x.Numero;
end;
end;
/

create table tb_quarto of tp_quarto(
Numero primary key,
Categoria not null,
capacidade not null
);

create or replace type tp_servico;
/

create or replace type tp_nt_ref_servico as table of ref tp_servico;
/

create or replace type tp_estadia as object(
Cadastro number,
Preco number(9,2),
Periodo varchar2(50),
Hospede REF tp_hospede,
Quarto REF tp_quarto,
Servicos_Realizados tp_nt_ref_servico,

member function getDataInicial return date,
member function getDataFinal return date 
);
/

create or replace type body tp_estadia as
member function getDataInicial return date is
begin
	return to_date(Substr(periodo, 1, 10), 'dd/mm/yyyy');
end;
member function getDataFinal return date is
begin
	return to_date(Substr(periodo, 23, 10), 'dd/mm/yyyy');
end;
end;
/

create or replace type tp_nt_ref_estadia as table of ref tp_estadia;
/

create table tb_estadia of tp_estadia(
Cadastro primary key,
Preco not null,
Periodo not null,
Hospede not null with rowid references tb_hospede,
Quarto not null with rowid references tb_quarto
)nested table Servicos_Realizados store as tb_servicos_realizados;

create or replace type tp_servico as object(
Cadastro number,
Tipo varchar2(50),
Nome varchar2(50),
Preco number(9,2),
Horario varchar2(50),
Lista_Estadias tp_nt_ref_estadia
);
/

create table servico of tp_servico(
Cadastro primary key
)nested table Lista_Estadias store as tb_lista_de_estadias;

create or replace type tp_fornecedor as object(
Cadastro number,
Nome varchar2(50),
Endereco varchar2(50),
CGC varchar2(50),
Razao_social varchar2(50)
);
/

create table tb_fornecedor of tp_fornecedor(
Cadastro primary key,
CGC unique
);

create or replace type tp_funcionario under tp_pessoa(
Salario number(9,2),
Departamento varchar2(50),
Funcao varchar2(50),
Servico REF tp_servico
);
/

create table tb_funcionario of tp_funcionario(
Cadastro primary key,
funcao not null,
salario not null,
servico not null with rowid references servico
);

create type tp_abastece as object(
Produto varchar2(50),
Data_Abastecimento date,
Cadastro_fornecedor ref tp_fornecedor,
Cadastro_servico ref tp_servico
);
/

create table tb_abastece of tp_abastece(
Produto not null,
Data_Abastecimento not null,
Cadastro_fornecedor not null with rowid references tb_fornecedor,
Cadastro_servico not null with rowid references servico
);

select deref(d) from table(select a.lista_servico from tb_estadia a where a.preco > 199.99) as d;

select * from table(select a.lista_servico from tb_estadia a where a.preco > 199.99);

select deref(value(d)) from table(select a.lista_servico from tb_estadia a where a.preco > 199.99) d; 

--Gerador de XML para a tabela tb_fornecedor
set serveroutput on;
declare
ctx dbms_xmlgen.ctxhandle;
result clob;
begin
-- criar um novo contexto para a consulta SQL
ctx := dbms_xmlgen.newContext('select * from tb_fornecedor');
-- personalizar a tag de entidade
DBMS_XMLGEN.setRowTag(ctx,'FORNECEDOR');
-- gerar um valor CLOB como resultado
result := dbms_xmlgen.getXML(ctx);
-- dar sada do resultado
dbms_output.put_line(result);
-- fechar o contexto
dbms_xmlgen.closeContext(ctx);
end;
/ 